# coding: utf-8
from __future__ import unicode_literals

from report_writer.core.report_objects.base import BaseReportObject
from report_writer.core.report_objects.cursor import Cursor
from report_writer.core.writers.base import BaseWriter


try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Any
    from typing import Tuple


class Simple(BaseReportObject):
    """
    Элементарный объект отчёта.

    Представляет собой единичную операцию записи, использующую область
    переданной высоты и ширины.

    Может быть использовано для клетки отчёта (тогда writer - писатель
    клетки отчёта).

    :param writer: Используемый писатель
    :param height: Высота области записи
    :param width: Ширина области записи
    """
    def __init__(self,
                 writer,  # type: BaseWriter
                 height=1,  # type: int
                 width=1,  # type: int
                 ):  # type: (...) -> None
        self._writer = writer  # type: BaseWriter
        self._height = height  # type: int
        self._width = width  # type: int
        # Если ширина и высота рывны 0, то операция не использует клетки отчёта
        self._use = (height, width) != (0, 0)  # type: bool

    @property
    def height(self
               ):  # type: () -> int
        """
        :return: Высота области записи писателя
        """
        return self._height

    @property
    def width(self
              ):  # type: () -> int
        """
        :return: Ширина области записи писателя
        """
        return self._width

    def write_object(self,
                     cursor,  # type: Cursor
                     data,  # type: Any
                     direction=Cursor.VERTICAL,  # type: Tuple[int, int]
                     ):  # type: (...) -> None
        with cursor.get_state() as state:
            if self._use:
                row, col = cursor.use_point()
                cursor.move(Cursor.VERTICAL, self.height - 1)
                cursor.move(Cursor.HORIZONTAL, self.width - 1)
                cursor.use_point()
            else:
                row, col = cursor.point

            self._writer.write(row, col, data)
            state.frame.close(direction)


# Псевдоним для более наглядного описания клеток отчёта
# (которые так же являются частным случаем элментарного объекта отчёта)
Cell = Simple
