# coding: utf-8
from __future__ import unicode_literals

from unittest import TestCase

from report_writer.core.report_objects.cursor import Cursor
from report_writer.core.report_objects.exceptions import (
    FrameOutOfBoundUsageException
)
from report_writer.core.report_objects.exceptions import NotLastStateUsage
from report_writer.core.report_objects.samples.iterators import (
    BorderedObjectIterator
)
from report_writer.core.report_objects.samples.iterators import (
    BorderedObjectIteratorWithHeaders
)
from report_writer.core.report_objects.samples.iterators import ObjectBlock
from report_writer.core.report_objects.samples.iterators import ObjectIterator
from report_writer.core.report_objects.samples.iterators import (
    ObjectIteratorWithHeaders
)
from report_writer.core.report_objects.samples.iterators import (
    SizedObjectIterator
)
from report_writer.core.report_objects.samples.simple import Cell
from report_writer.utils.mock_report import CellWriter
from report_writer.utils.mock_report import FormattedCellWriter
from report_writer.utils.mock_report import FormattedMergedCellWriter
from report_writer.utils.mock_report import MergedCellWriter
from report_writer.utils.mock_report import NamedFormatWriter
from report_writer.utils.mock_report import ReportFileMock


class TestReportIterableObjects(TestCase):
    def test_object_iterator(self):
        report = ReportFileMock(5, 2)
        obj = ObjectIterator(Cell(CellWriter(report)), Cursor.VERTICAL)

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, range(5), Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [0, None],
            [1, None],
            [2, None],
            [3, None],
            [4, None],
        ])

    def test_object_iterator_cursor(self):
        report = ReportFileMock(5, 2)
        obj = ObjectIterator(Cell(CellWriter(report)), Cursor.VERTICAL)

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, range(5), Cursor.HORIZONTAL)
            obj.write_object(state.cursor, range(5), Cursor.HORIZONTAL)

        self.assertEqual(report.data, [
            [0, 0],
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
        ])

    def test_object_iterator_first_last(self):
        report = ReportFileMock(6, 2)

        obj = BorderedObjectIterator(
            Cell(CellWriter(report)), Cursor.VERTICAL,
            first_object=Cell(FormattedCellWriter(report, 'first')),
            last_object=Cell(FormattedCellWriter(report, 'last')),
            single_object=Cell(FormattedCellWriter(report, 'single')),
        )

        with Cursor(0, 0).get_state() as state:
            NamedFormatWriter(report, 'first', 0).write()
            NamedFormatWriter(report, 'last', 1).write()
            NamedFormatWriter(report, 'single', 2).write()

            obj.write_object(state.cursor, [
                (0, True, False),
                (1, False, False),
                (2, False, False),
                (3, False, False),
                (4, False, True),
            ], Cursor.VERTICAL)
            obj.write_object(state.cursor, [
                (0, True, True),
            ], Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [(0, 0), None],
            [1,      None],
            [2,      None],
            [3,      None],
            [(1, 4), None],
            [(2, 0), None],
        ])

    def test_object_block(self):
        report = ReportFileMock(3, 3)
        cell = Cell(CellWriter(report))
        merged = Cell(MergedCellWriter(report, 1, 2), 1, 2)
        obj = ObjectBlock([cell, merged], Cursor.VERTICAL)

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [0, 1], Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [0,    None, None],
            [1,    None, None],
            [None, None, None],
        ])

        self.assertIn({(1, 0), (1, 1)}, report.merged)

    def test_object_block_cursor(self):
        report = ReportFileMock(3, 5)
        cell = Cell(CellWriter(report))
        merged = Cell(MergedCellWriter(report, 1, 2), 1, 2)
        obj = ObjectBlock([cell, merged], Cursor.VERTICAL)

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [0, 1], Cursor.HORIZONTAL)
            obj.write_object(state.cursor, [2, 3], Cursor.HORIZONTAL)

        self.assertEqual(report.data, [
            [0,    None, 2,    None, None],
            [1,    None, 3,    None, None],
            [None, None, None, None, None],
        ])

        self.assertIn({(1, 0), (1, 1)}, report.merged)
        self.assertIn({(1, 2), (1, 3)}, report.merged)

    def test_sized_object_iterator(self):
        report = ReportFileMock(3, 3)
        cell = Cell(CellWriter(report))
        obj = SizedObjectIterator(cell, Cursor.VERTICAL, 2)

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [1, 2, 3], Cursor.HORIZONTAL)

        self.assertEqual(report.data, [
            [1,    None, None],
            [2,    None, None],
            [None, None, None],
        ])

        report = ReportFileMock(3, 3)
        cell = Cell(CellWriter(report))
        obj = SizedObjectIterator(cell, Cursor.HORIZONTAL, 2)

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [1, 2, 3], Cursor.HORIZONTAL)

        self.assertEqual(report.data, [
            [1,    2,    None],
            [None, None, None],
            [None, None, None],
        ])

    def test_sized_object_iterator_cursor(self):
        report = ReportFileMock(3, 3)
        cell = Cell(CellWriter(report))
        obj = SizedObjectIterator(cell, Cursor.VERTICAL, 2)

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [1, 2, 3], Cursor.HORIZONTAL)
            obj.write_object(state.cursor, [1, 2, 3], Cursor.HORIZONTAL)

        self.assertEqual(report.data, [
            [1,    1,    None],
            [2,    2,    None],
            [None, None, None],
        ])


class TestGroupedObjectIterator(TestCase):
    def test_grouped_object_iterator(self):
        report = ReportFileMock(7, 3)
        cell = Cell(CellWriter(report))
        group = ObjectIterator(cell, Cursor.VERTICAL)
        obj = ObjectIteratorWithHeaders(
            group, ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(MergedCellWriter(report, h, w), h, w),
                lambda: Cell(CellWriter(report))),
            Cursor.VERTICAL,
        )

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [
                (1, [1, 2, 3]),
                (2, [1, 2]),
                (3, [1]),
            ], Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [1,    1,    None],
            [None, 2,    None],
            [None, 3,    None],
            [2,    1,    None],
            [None, 2,    None],
            [3,    1,    None],
            [None, None, None],
        ])

        self.assertIn({(0, 0), (1, 0), (2, 0)}, report.merged)
        self.assertIn({(3, 0), (4, 0)}, report.merged)
        self.assertNotIn({(5, 0)}, report.merged)

    def test_grouped_object_border(self):
        report = ReportFileMock(14, 3)
        cell = Cell(CellWriter(report))
        group = ObjectIterator(cell, Cursor.VERTICAL)
        obj = BorderedObjectIteratorWithHeaders(
            group, ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(MergedCellWriter(report, h, w), h, w),
                lambda: Cell(CellWriter(report))),
            Cursor.VERTICAL,
            first_header_object=ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(
                    FormattedMergedCellWriter(report, h, w, 'first'), h, w),
                lambda: Cell(FormattedCellWriter(report, 'first'))),
            last_header_object=ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(
                    FormattedMergedCellWriter(report, h, w, 'last'), h, w),
                lambda: Cell(FormattedCellWriter(report, 'last'))),
            single_header_object=ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(
                    FormattedMergedCellWriter(report, h, w, 'single'), h, w),
                lambda: Cell(FormattedCellWriter(report, 'single'))),
        )

        with Cursor(0, 0).get_state() as state:
            NamedFormatWriter(report, 'first', 0).write()
            NamedFormatWriter(report, 'last', 1).write()
            NamedFormatWriter(report, 'single', 2).write()
            obj.write_object(state.cursor, [
                ((1, [1, 2, 3]), True, False),
                ((2, [1, 2]), False, False),
                ((3, [1, 2]), False, True),
            ], Cursor.VERTICAL)
            obj.write_object(state.cursor, [
                ((1, [1, 2]), True, True),
            ], Cursor.VERTICAL)
            obj.write_object(state.cursor, [
                ((1, []), True, True)
            ], Cursor.VERTICAL)
            obj.write_object(state.cursor, [
                ((1, [1]), True, False),
                ((2, [1]), False, False),
                ((3, [1]), False, True),
            ], Cursor.VERTICAL)
            obj.write_object(state.cursor, [
                ((1, [1]), True, True),
            ], Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [(0, 1), 1,    None],
            [None,   2,    None],
            [None,   3,    None],
            [2,      1,    None],
            [None,   2,    None],
            [(1, 3), 1,    None],
            [None,   2,    None],
            [(2, 1), 1,    None],
            [None,   2,    None],
            [(0, 1), 1,    None],
            [2,      1,    None],
            [(1, 3), 1,    None],
            [(2, 1), 1,    None],
            [None,   None, None]
        ])

    def test_grouped_object_inner_group(self):
        report = ReportFileMock(11, 4)
        cell = Cell(CellWriter(report))
        group = ObjectIterator(cell, Cursor.VERTICAL)
        inner = ObjectIteratorWithHeaders(
            group, ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(MergedCellWriter(report, h, w), h, w),
                lambda: Cell(CellWriter(report))),
            Cursor.VERTICAL
        )
        obj = ObjectIteratorWithHeaders(
            inner, ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(MergedCellWriter(report, h, w), h, w),
                lambda: Cell(CellWriter(report))),
            Cursor.VERTICAL,
        )

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [
                (1, (
                    (1, [1, 2, 3]),
                    (2, [1, 2]),
                )),
                (2, (
                    (1, [1, 2, 3]),
                    (2, [1]),
                )),
                (3, (

                )),
                (4, (
                    (1, [1]),
                ))
            ], Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [1,    1,    1,    None],
            [None, None, 2,    None],
            [None, None, 3,    None],
            [None, 2,    1,    None],
            [None, None, 2,    None],
            [2,    1,    1,    None],
            [None, None, 2,    None],
            [None, None, 3,    None],
            [None, 2,    1,    None],
            [4,    1,    1,    None],
            [None, None, None, None],
        ])

        self.assertIn({(0, 0), (1, 0), (2, 0), (3, 0), (4, 0)}, report.merged)
        self.assertIn({(5, 0), (6, 0), (7, 0), (8, 0)}, report.merged)
        self.assertIn({(0, 1), (1, 1), (2, 1)}, report.merged)
        self.assertIn({(3, 1), (4, 1)}, report.merged)
        self.assertIn({(5, 1), (6, 1), (7, 1)}, report.merged)
        self.assertNotIn({(8, 1)}, report.merged)

        report = ReportFileMock(7, 7)
        cell = Cell(CellWriter(report))
        group = ObjectIterator(cell, Cursor.VERTICAL)
        inner = ObjectIteratorWithHeaders(
            group, ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(MergedCellWriter(report, h, w), h, w),
                lambda: Cell(CellWriter(report))),
            Cursor.VERTICAL
        )

        obj = ObjectIteratorWithHeaders(
            inner, ObjectIteratorWithHeaders.make_header_factory(
                lambda h, w: Cell(MergedCellWriter(report, h, w), h, w),
                lambda: Cell(CellWriter(report))  # pragma: no cover
            ),
            Cursor.HORIZONTAL
        )

        with Cursor(0, 0).get_state() as state:
            obj.write_object(state.cursor, [
                (1, (
                    (1, [1, 2, 3]),
                    (2, [1, 2]),
                )),
                (2, (
                    (1, [1, 2, 3]),
                    (2, [1]),
                )),
                (3, (

                )),
                (4, (
                    (1, [1]),
                ))
            ], Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [1,    None, 2,    None, 4,    None, None],
            [1,    1,    1,    1,    1,    1,    None],
            [None, 2,    None, 2,    None, None, None],
            [None, 3,    None, 3,    None, None, None],
            [2,    1,    2,    1,    None, None, None],
            [None, 2,    None, None, None, None, None],
            [None, None, None, None, None, None, None],
        ])

        self.assertIn({(0, 0), (0, 1)}, report.merged)
        self.assertIn({(0, 2), (0, 3)}, report.merged)
        self.assertIn({(1, 0), (2, 0), (3, 0)}, report.merged)
        self.assertIn({(1, 2), (2, 2), (3, 2)}, report.merged)
        self.assertIn({(4, 0), (5, 0)}, report.merged)
        self.assertNotIn({(4, 2)}, report.merged)


class TestCursor(TestCase):
    def test_cursor_move(self):
        c = Cursor(0, 0)

        with c.get_state() as state:
            self.assertEqual(state.point, (0, 0))

            state.cursor.move(Cursor.HORIZONTAL, 1)
            self.assertEqual(c.point, (0, 1))
            self.assertEqual(state.point, (0, 0))

            state.cursor.back()

            state.cursor.move(Cursor.VERTICAL, 1)
            self.assertEqual(c.point, (1, 0))
            self.assertEqual(state.point, (0, 0))

            state.cursor.back()

            state.cursor.move(Cursor.HORIZONTAL, 3)
            self.assertEqual(c.point, (0, 3))
            self.assertEqual(state.point, (0, 0))

            state.cursor.back()

            state.cursor.move(Cursor.VERTICAL, 3)
            self.assertEqual(c.point, (3, 0))
            self.assertEqual(state.point, (0, 0))

            state.cursor.back()

    def test_cursor_state(self):
        c = Cursor(3, 3)

        with c.get_state() as state:
            state.cursor.move(Cursor.VERTICAL, 3)
            state.cursor.move(Cursor.HORIZONTAL, 2)

            self.assertEqual(state.cursor.use_point(), (6, 5))

            self.assertEqual(state.frame.width, 3)
            self.assertEqual(state.frame.height, 4)

            with c.get_state() as state0:
                state.cursor.move(Cursor.VERTICAL, 2)
                state.cursor.move(Cursor.HORIZONTAL, 3)

                self.assertEqual(state.cursor.use_point(), (8, 8))

                self.assertEqual(state.frame.width, 6)
                self.assertEqual(state.frame.height, 6)

                self.assertEqual(state0.frame.width, 4)
                self.assertEqual(state0.frame.height, 3)

                state0.cursor.back()
                self.assertEqual(c.point, (6, 5))

            state.cursor.back()
            self.assertEqual(c.point, (3, 3))

    def test_cursor_out_of_frame(self):
        c = Cursor(0, 0)

        with c.get_state() as state:
            state.cursor.move(Cursor.HORIZONTAL, 1)
            with c.get_state() as _:
                with self.assertRaises(NotLastStateUsage):
                    c.back(state)

                with self.assertRaises(FrameOutOfBoundUsageException):
                    c.move((-1, 0), 1)
                    c.use_point()

    def test_cursor_state_closing(self):
        c = Cursor(0, 0)

        with self.assertRaises(NotLastStateUsage):
            # При выходе из менеджера попытается удалить не последнее
            # состояние, т.к. внутри менеджера мы добавили ещё одно
            with c.get_state() as _:
                state0 = c._create_state_point()
                c.add_state_point(state0)

        c0 = Cursor(0, 0)
        c1 = Cursor(0, 0)

        with self.assertRaises(ValueError):
            c0.add_state_point(c1._create_state_point())

        with self.assertRaises(ValueError):
            c0.remove_state_point(c1._create_state_point())

        c = Cursor(0, 0)
        with c.get_state() as state:
            c.move(Cursor.VERTICAL, 2)
            c.move(Cursor.HORIZONTAL, 2)

            c.use_point()

            self.assertEqual(state.frame.height, 3)
            self.assertEqual(state.frame.width, 3)

            state.frame.close(Cursor.VERTICAL)
            self.assertEqual(c.point, (3, 0))

        c = Cursor(0, 0)
        with c.get_state() as state:
            c.move(Cursor.VERTICAL, 2)
            c.move(Cursor.HORIZONTAL, 2)

            c.use_point()

            state.frame.close(Cursor.HORIZONTAL)
            self.assertEqual(c.point, (0, 3))

        c = Cursor(0, 0)
        with c.get_state() as state:
            c.move(Cursor.VERTICAL, 2)
            c.move(Cursor.HORIZONTAL, 2)

            c.use_point()

            state.frame.reset_frame()
            state.frame.close(Cursor.HORIZONTAL)
            self.assertEqual(c.point, (0, 1))

    def test_cursor_invert_direction(self):
        self.assertEqual(
            Cursor.invert_direction(Cursor.VERTICAL), Cursor.HORIZONTAL)
        self.assertEqual(
            Cursor.invert_direction(Cursor.HORIZONTAL), Cursor.VERTICAL)
